import itertools
import re


class PVar:
    p = re.compile('^(.+)-(\d+)$|^(.+)$')

    def __init__(self, var: str, i: int = 0):
        self.var = var
        self.i = i

    def _tuple_for_ordering(self):
        return (self.i, self.var)

    def __eq__(self, other):
        return self._tuple_for_ordering().__eq__(other._tuple_for_ordering())
        # return self.var == other.var and self.i == other.i

    def __lt__(self, other):
        return self._tuple_for_ordering().__lt__(other._tuple_for_ordering())

    def __le__(self, other):
        return self._tuple_for_ordering().__le__(other._tuple_for_ordering())

    def __gt__(self, other):
        return self._tuple_for_ordering().__gt__(other._tuple_for_ordering())

    def __ge__(self, other):
        return self._tuple_for_ordering().__ge__(other._tuple_for_ordering())

    @classmethod
    def from_(cls, x):
        if isinstance(x, cls):
            return cls(x.var, x.i)  # make a copy of the Key object
        elif isinstance(x, tuple):
            return cls(*x)  # assume it's a (var, i) tuple
        elif isinstance(x, str):
            return cls.from_str(x)
        else:
            return cls(*x.__iter__())

    def __iter__(self):
        return self.var, self.i

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.var}', {self.i})"

    def __hash__(self):
        return hash(self.__repr__())

    def __str__(self):
        if self.i == 0:
            return f"{self.var}"
        else:
            return f"{self.var}{self.i}"

    @classmethod
    def from_str(cls, s):
        s = s.strip()
        m = cls.p.match(s)
        g = m.groups()
        if g[-1] is None:
            return cls(var=g[0], i=-int(g[1]))
        else:
            return cls(var=g[-1])

    def __getitem__(self, i):
        return PVar(self.var, self.i + i)

    def __add__(self, i):
        return PVar(self.var, self.i + i)

    def __sub__(self, i):
        return PVar(self.var, self.i - i)

    def __mul__(self, other):
        return VarSet([self, other])


class VarSet:
    def __init__(self, *varset):
        if len(varset) == 1 and isinstance(varset[0], (tuple, list, VarSet)):
            varset = varset[0]
        varset = list(map(PVar.from_, varset))
        self.varset = sorted(varset)
        self.min_abs_i = abs(min(x.i for x in self))  # TODO: Not enough: Need to check on upper bound of sliding win

    @property
    def varset_strs(self):
        return list(map(str, self.varset))

    def __mul__(self, other):
        if isinstance(other, PVar):
            return VarSet(self.varset + [other])
        else:
            return VarSet(self.varset + other)

    def __eq__(self, other):
        if len(self.varset) != len(other.varset):
            return False
        else:
            for k, kk in zip(self.varset, other.varset):
                if k != kk:
                    return False
            return True

    def __iter__(self):
        return iter(self.varset)

    def __repr__(self):
        s = ", ".join(map(lambda x: "'{}'".format(x), self.varset))
        return f"{self.__class__.__name__}({s})"

    def __str__(self):
        return "(" + ", ".join(map(str, self.varset)) + ")"

    def __hash__(self):
        return hash(self.__repr__())

    def __getitem__(self, i):
        varset = [k[i] for k in self.varset]
        return VarSet(varset)


class VarSetFactory:
    @staticmethod
    def single_dim_markovs(varnames):
        return map(lambda v: VarSet([(v, -1), (v, 0)]), varnames)

    @staticmethod
    def pairs(varnames):
        return map(VarSet, itertools.combinations(map(PVar, varnames), 2))

    @staticmethod
    def tuples(varnames, tuple_size: int = 2):
        return map(VarSet, itertools.combinations(map(PVar, varnames), tuple_size))

    @staticmethod
    def markov_pairs(varnames):
        return map(lambda v: VarSet(PVar(v[0], -1), PVar(v[1], 0)), itertools.product(varnames, varnames))

    @staticmethod
    def from_str(s):
        return VarSet(*map(PVar.from_str, list(s[1:-1].split(','))))

    # @classmethod
    # def pairs(cls, vars):
    #     return list(itertools.combinations(vars, 2))


def extract_kps(df, kps):
    # keep only elements of kps that have columns for them
    cols = set(df.columns)
    _kps = list()
    for k in kps:
        if k.var in cols:
            _kps.append(k)

    for i in range(len(df)):
        if i >= kps.min_abs_i:
            yield tuple(df[k.var].iloc[i + k.i] for k in _kps)


class DfData:
    def __init__(self, df):
        self.df = df

    def __getitem__(self, k):
        if isinstance(k, PVar):
            return self.df[k.var].iloc[k.i]
        elif isinstance(k, VarSet):
            return tuple(map(self.__getitem__, k))

    def extract_with_key_pattern_sets(self, kps_list):
        for i in range(len(self.df)):
            for kps in kps_list:
                if i >= kps.min_abs_i:
                    yield kps, self[kps[i]]

    def extract_kps(self, kps):
        for i in range(len(self.df)):
            if i >= kps.min_abs_i:
                yield self[kps[i]]



import matplotlib.pylab as plt
from itertools import islice
from typing import Iterable
import numpy as np
import pandas as pd

from odus.util import write_images

ihead = lambda it: islice(it, 0, 5)  # just a little useful util. Not used within module


def heatmap(X, y=None, col_labels=None, figsize=None, cmap=None, return_gcf=False, ax=None,
            xlabel_top=True, ylabel_left=True, xlabel_bottom=True, ylabel_right=True, **kwargs):
    n_items, n_cols = X.shape
    if col_labels is not None:
        if col_labels is not False:
            assert len(col_labels) == n_cols, \
                "col_labels length should be the same as the number of columns in the matrix"
    elif isinstance(X, pd.DataFrame):
        col_labels = list(X.columns)

    if figsize is None:
        x_size, y_size = X.shape
        if x_size >= y_size:
            figsize = (6, min(18, 6 * x_size / y_size))
        else:
            figsize = (min(18, 6 * y_size / x_size), 6)

    if cmap is None:
        if X.min(axis=0).min(axis=0) < 0:
            cmap = 'RdBu_r'
        else:
            cmap = 'hot_r'

    kwargs['cmap'] = cmap
    kwargs = dict(kwargs, interpolation='nearest', aspect='auto')

    if figsize is not False:
        plt.figure(figsize=figsize)

    if ax is None:
        plt.imshow(X, **kwargs)
    else:
        ax.imshow(X, **kwargs)
    plt.grid(None)

    if y is not None:
        y = np.array(y)
        assert all(sorted(y) == y), "This will only work if your row_labels are sorted"

        unik_ys, unik_ys_idx = np.unique(y, return_index=True)
        for u, i in zip(unik_ys, unik_ys_idx):
            plt.hlines(i - 0.5, 0 - 0.5, n_cols - 0.5, colors='b', linestyles='dotted', alpha=0.5)
        plt.hlines(n_items - 0.5, 0 - 0.5, n_cols - 0.5, colors='b', linestyles='dotted', alpha=0.5)
        plt.yticks(unik_ys_idx + np.diff(np.hstack((unik_ys_idx, n_items))) / 2, unik_ys)
    elif isinstance(X, pd.DataFrame):
        y_tick_labels = list(X.index)
        plt.yticks(list(range(len(y_tick_labels))), y_tick_labels);

    if col_labels is not None:
        plt.xticks(list(range(len(col_labels))), col_labels)
    else:
        plt.xticks([])

    plt.gca().xaxis.set_tick_params(labeltop=xlabel_top, labelbottom=xlabel_bottom)
    plt.gca().yaxis.set_tick_params(labelleft=ylabel_left, labelright=ylabel_right)

    if return_gcf:
        return plt.gcf()


def plot_life_course(df, grid=False, figsize=3.5, **kwargs):
    if isinstance(figsize, (int, float)):  # if figsize is a number, it's a factor of the df size (shape)
        figsize = np.array(df.shape) / figsize
    kwargs['figsize'] = figsize
    heatmap(df.T, **kwargs);
    plt.grid(grid);


def plot_life(df, fields=None, title=None, ax=None):
    if fields is None:
        fields = slice(None, None)
    plot_life_course(df[fields], ax=ax)
    plt.grid(which='both', axis='x')
    if title is not None:
        plt.title(title)
        plt.gca().xaxis.set_tick_params(labeltop=False, labelbottom=True)


def _get_keys(df_store, keys) -> Iterable:
    if keys is None:
        keys = df_store.keys()
    elif callable(keys):
        keys_filt = keys
        keys = filter(keys_filt, df_store.keys())
    assert isinstance(keys, Iterable), "keys should be iterable at this point"
    return keys


def life_plots(df_store, fields=None, keys=None, k_df_to_title=lambda k, df: k.split('/')[1]):
    keys = _get_keys(df_store, keys)
    for k in keys:
        df = df_store[k]
        plot_life(df, fields, title=k_df_to_title(k, df))
        yield plt.gca()


def write_trajectories_to_file(df_store, fields=None, keys=None, fp='test.pdf', pil_write_format=None,
                               to_pil_image_kwargs=None, **pil_save_kwargs):
    keys = _get_keys(df_store, keys)

    def figs():
        fig_gen = map(lambda k, df: plot_life(df, fields, title=k.split('/')[1]),
                      *zip(*((k, df_store[k]) for k in keys)))
        for _ in fig_gen:
            fig = plt.gcf()
            yield fig
            fig.clear()

    write_images(figs(), fp, pil_write_format, to_pil_image_kwargs, **pil_save_kwargs)



from odus.pobj import add_method

########################################################################################################################
# If you want an element that is neutral for any type and any method (well, all non-operator methods, plus a few ops)...

# TODO: Added __add__, __mul__, etc. because __getattribute__ doesn't work with operator overloading
# TODO: ... (e.g. works for __add__ works, but not with +).


class GreatUniversalNothing(object):
    """
    An object that is a neutral for any method.
    Only works with some operators (e.g. works for __add__ works, but not with +).

    >>> nothing = GreatUniversalNothing()  # note, this is constructed ones, and used many times
    >>>
    >>> nothing + 2
    2
    >>> 2 + nothing
    2
    >>> nothing + 2 + nothing
    2
    >>> nothing * 3.14 * nothing
    3.14
    >>> nothing | [1, 2, 3] | nothing
    [1, 2, 3]
    >>> nothing & 'hi there' & nothing
    'hi there'
    >>> # Now we demo how we get the same behavior no matter what the type is
    >>> nothing + 2
    2
    >>> nothing + [1, 2, 3]
    [1, 2, 3]
    >>> print((nothing + "wassup world!"))
    wassup world!
    >>>
    >>> # Useful as an alternative of using zeros(shape_of_your_array) to initialized array cumulators.
    >>> # No need to specify the array shape in advance!
    >>> import numpy as np
    >>> print((nothing + np.array([1, 2, 3, 4])))
    [1 2 3 4]
    >>> # and += works automatically!
    >>> t = nothing
    >>> t += np.array([1, 2, 3, 4])
    >>> print(t)
    [1 2 3 4]
    >>> t += np.array([10, 20, 30, 40])
    >>> print(t)
    [11 22 33 44]
    >>>
    >>> # Example with defaultdict (and multi type value cumulators
    >>> from collections import defaultdict
    >>> t = defaultdict(lambda: nothing)
    >>> t['foo'] += 2
    >>> t['bar'] += [1,2,3]
    >>> t['bar'] += [10,20]
    >>> t['is'] += np.array([1,2,3,4])
    >>> t['is'] += np.array([1,2,3,4])
    >>> t['overused'] += "hello"
    >>> t['overused'] += " "
    >>> t['overused'] += "world"
    >>> dict(t)
    {'overused': 'hello world', 'is': array([2, 4, 6, 8]), 'foo': 2, 'bar': [1, 2, 3, 10, 20]}
    >>>
    >>> # Works with other methods too!
    >>> nothing * 314
    314
    >>> nothing * 'foobar'
    'foobar'
    """

    def __getattribute__(self, name):
        return lambda x: x

    def __bool__(self):
        return False

    def __add__(self, other):
        return other

    def __radd__(self, other):
        return other

    def __mul__(self, other):
        return other

    def __rmul__(self, other):
        return other

    def __or__(self, other):
        return other

    def __ror__(self, other):
        return other

    def __and__(self, other):
        return other

    def __rand__(self, other):
        return other


nothing = GreatUniversalNothing()  # most of the time, this is the only thing you need to import


########################################################################################################################
# If you want an element that is neutral for any type but only a specific method


def transparent_method(self, x):
    return x


def get_universal_neutral_for_method(method_name):
    """
    A universal neutral element (synonyms: "identity element" and "unit element").
    A neutral element is with respect to a binary operation op(a, b).
    A neutral element e (for op) is one such that op(e, b) = b for any b.
    (Formally, we should also have op(b, e) = b, but we'll ignore that!)

    It is universal from the point of view of type, not method.
    Here, you specify what method (operation) you want it to be a neutral element for.
    Typically, you use this when you want to initialize a variable that will "accumulate/aggregate" objects one by one.
    For example, you say
        cumul = 0
    and then add elements to cumul
        cumul = cumul + x
    or
        cumul += x
    etc.

    Depending on what the type is, you need to initialize in different ways.
        cumul = 0   for ints
        cumul = []  for lists
        cumul = numpy.array(array_shape) for numpy arrays (and better know the dims of the arrays you'll be feeding!)
        cumul = "" for strings

    With the universal element, you don't need to initialize according to type. You just need to say:
        cumul = get_universal_neutral_for_method("__add__")
    and you'll have something that behaves as expected: Be transparent.

    Other examples of common methods where you want a neutral:  __mul__, update, etc.

    >>> nothing = get_universal_neutral_for_method('__add__')  # note, this is constructed ones, and used many times
    >>>
    >>> nothing + 2
    2
    >>> nothing + [1, 2, 3]
    [1, 2, 3]
    >>> print((nothing + "wassup world!"))
    wassup world!
    >>>
    >>> # Useful as an alternative of using zeros(shape_of_your_array) to initialized array cumulators.
    >>> # No need to specify the array shape in advance!
    >>> import numpy as np
    >>> print((nothing + np.array([1, 2, 3, 4])))
    [1 2 3 4]
    >>> # and += works automatically!
    >>> t = nothing
    >>> t += np.array([1, 2, 3, 4])
    >>> print(t)
    [1 2 3 4]
    >>> t += np.array([10, 20, 30, 40])
    >>> print(t)
    [11 22 33 44]
    >>>
    >>> # Example with defaultdict (and multi type value cumulators
    >>> from collections import defaultdict
    >>> t = defaultdict(lambda: nothing)
    >>> t['foo'] += 2
    >>> t['bar'] += [1,2,3]
    >>> t['bar'] += [10,20]
    >>> t['is'] += np.array([1,2,3,4])
    >>> t['is'] += np.array([1,2,3,4])
    >>> t['overused'] += "hello"
    >>> t['overused'] += " "
    >>> t['overused'] += "world"
    >>> dict(t)
    {'overused': 'hello world', 'is': array([2, 4, 6, 8]), 'foo': 2, 'bar': [1, 2, 3, 10, 20]}
    >>>
    >>> # Works with other methods too!
    >>> nothing = get_universal_neutral_for_method('__mul__')
    >>> nothing * 314
    314
    >>> nothing * 'foobar'
    'foobar'
    """
    return add_method(object(), method_func=transparent_method, method_name=method_name,
                      class_name=method_name + '_UniversalNeutral')


addition_neutral = get_universal_neutral_for_method('__add__')
multiplication_neutral = get_universal_neutral_for_method('__mul__')
update_neutral = get_universal_neutral_for_method('update')


########################################################################################################################
# If you want an element that is neutral for any type and an explicit list of methods

def get_universal_neutral_for_methods(method_names):
    """
    See get_universal_neutral_for_method(method_name).
    This is just the "multiple methods" version of that one.
    """
    nothing = object
    for method_name in method_names:
        nothing = add_method(nothing, method_func=transparent_method, method_name=method_name,
                             class_name='UniversalNeutral')
    return nothing





from itertools import chain
import types


def add_method(obj, method_func, method_name=None, class_name=None):
    """
    Dynamically add a method to an object.

    :param obj: The object to add a method to
    :param method_func: The function to use as a method. The first argument must be the object itself
        (usually called self)
    :param method_name: The desired function name. If None, will take method_func.__name__
    :param class_name:  The desired class name. If None, will take type(obj).__name__
    :return: the object, but with the additional method (or a different function for it)

    >>> class A:
    ...     def __init__(self, x=10):
    ...         self.x = x
    >>> def times(self, y):
    ...     return self.x * y
    >>> def plus(self, y):
    ...     return self.x + y
    >>> a = A(x=10)
    >>> a = add_method(a, plus, '__call__')  # add a __call__ method, assigning it to plus
    >>> a(2)
    12
    >>> a = add_method(a, times, '__call__')  # reassign the __call__ method to times instead
    >>> a(2)
    20
    >>> a = add_method(a, plus, '__getitem__')  # assign the method __getitem__ to plus
    >>> a[2]  # see that it works
    12
    >>> a(2)  # and that we still have our __call__ method
    20
    """
    if isinstance(method_func, str):
        method_name = method_func
        method_func = getattr(obj, method_name)
    if method_name is None:
        method_name = method_func.__name__

    base = type(obj)

    if class_name is None:
        class_name = base.__name__
    bases = (base.__bases__[1:]) + (base,)
    bases_names = set(map(lambda x: x.__name__, bases))
    if class_name in bases_names:
        for i in range(6):
            class_name += '_'
            if not class_name in bases_names:
                break
        else:
            raise ValueError("can't find a name for class that is not taken by bases. Consider using explicit name")

    new_keys = set(dir(obj)) - set(chain(*[dir(b) for b in bases]))

    d = {a: getattr(obj, a) for a in new_keys}
    d[method_name] = method_func

    return type(class_name, bases, d)()


def inject_method(obj, method_func, method_name=None):
    """
    Inject one or several methods to an object.
    Note that it doesn't work if method_name is a dunder.
    If you want to inject things like __call__, __getitem__, etc., use add_method instead.
    :param obj: The object to add a method to
    :param method_func: The function to use as a method. The first argument must be the object itself
        (usually called self)
    :param method_name: The desired function name. If None, will take method_func.__name__
    :return: the object, but with the additional method (or a different function for it)

    >>> class A:
    ...     def __init__(self, x=10):
    ...         self.x = x
    >>> def times(self, y):
    ...     return self.x * y
    >>> def plus(self, y):
    ...     return self.x + y
    >>> a = A(x=10)
    >>> a = inject_method(a, plus, 'add')  # inject the plus method (but call it 'add')
    >>> a.add(2)
    12
    >>> a = inject_method(a, times)  # inject the times method (and use it's name automatically)
    >>> a.times(2)
    20
    >>> a = inject_method(a, plus, 'times')  # be evil: Inject the plus method but call it 'times'
    >>> a.times(2)
    12
    """
    if isinstance(method_func, types.FunctionType):
        if method_name is None:
            method_name = method_func.__name__
        setattr(obj,
                method_name,
                types.MethodType(method_func, obj))
    else:
        if isinstance(method_func, dict):
            method_func = [(func, func_name) for func_name, func in method_func.items()]
        for method in method_func:
            if isinstance(method, tuple) and len(method) == 2:
                obj = inject_method(obj, method[0], method[1])
            else:
                obj = inject_method(obj, method)

    return obj



from os.path import dirname, join, sep
import pandas as pd
import numpy as np
import matplotlib.pylab as plt
from itertools import combinations, chain

from odus.sequential_var_sets import PVar, VarSet, DfData, VarSetFactory
from spyn.ppi.pot import Pot
from odus.dacc import (
    DfStore,
    counts_of_kps,
    Dacc,
    VarSetCountsStore,
    mk_pvar_struct,
    PotStore,
    _commun_columns_of_dfs,
    Struct,
    mk_pvar_str_struct,
    VarStr,
)
from odus.plot_utils import (
    plot_life_course,
    plot_life,
    life_plots,
    write_trajectories_to_file,
)

# __all__ = ['plot_life_course', 'plot_life', 'life_plots', 'write_trajectories_to_file',
#            'counts_of_kps', 'Dacc', 'VarSetCountsStore', 'PVar', 'VarSet', 'DfData', 'VarSetFactory']

package_dir = dirname(__file__)
DFLT_SURVEY_DIR = join(
    package_dir,
)
dflt_figsize = (16, 5)


def get_stores_v_and_s(survey_dir=DFLT_SURVEY_DIR):
    df_store = DfStore(survey_dir + sep + '{}.xlsx', mode='b')
    #     cstore = VarSetCountsStore(df_store)
    pstore = PotStore(df_store)
    v = mk_pvar_struct(df_store, only_for_cols_in_all_dfs=True)
    s = mk_pvar_str_struct(v)
    return df_store, pstore, v, s


def get_cstores_v_and_s(survey_dir=DFLT_SURVEY_DIR):
    df_store = DfStore(survey_dir + sep + '{}.xlsx', mode='b')
    pstore = PotStore(df_store)
    v = mk_pvar_struct(df_store, only_for_cols_in_all_dfs=True)
    s = mk_pvar_str_struct(v)
    return pstore, v, s


def get_markov_rel_risk(pstore, fields=None):
    if fields is None:
        fields = _commun_columns_of_dfs(pstore.df_store.values())
    c = list()
    for vs in VarSetFactory.markov_pairs(fields):
        p = pstore[vs]
        if vs.varset[0].i == -1:
            exposure = str(vs.varset[0])
            event = str(vs.varset[1])
        else:
            event = str(vs.varset[0])
            exposure = str(vs.varset[1])
        rel_risk = p.relative_risk(
            event_var=event, exposure_var=exposure, event_val=1, exposure_val=1
        )

        c.append({'exposure': exposure, 'event': event, 'rel_risk': rel_risk})

    markov_rel_risk = pd.DataFrame(c)
    return markov_rel_risk.pivot(index='exposure', columns='event', values='rel_risk')


def remission_relative_risk(pstore, event, exposure):
    """
    Returns the remission influence of exposure on event.
    By "remission influence" we mean the relative risk of the event, given the exposure to an influencer
    during the "previous" time tick, conditioned on the existence of the event.
    That is, we're looking at:
        P(event=true | exposure-1=true, event-1=true)
    divided by
        P(event=true | exposure-1=false, event-1=true)
    """
    if isinstance(event, str):
        event = VarStr(event)
    if isinstance(exposure, str):
        exposure = VarStr(exposure)
    p = pstore[exposure - 1, event - 1, event]
    event_last_year = p * Pot.from_hard_evidence(**{event - 1: 1}) >> [
        exposure - 1,
        event,
    ]
    return event_last_year.relative_risk(event, exposure - 1, smooth_count=1)


def get_markov_remission_rel_risk(pstore, fields=None):
    if fields is None:
        fields = _commun_columns_of_dfs(pstore.df_store.values())
    c = list()
    for event, exposure in chain(
        *([(x, y), (y, x)] for x, y in combinations(fields, 2))
    ):
        rel_risk = remission_relative_risk(pstore, event, exposure)
        c.append({'exposure': exposure, 'event': event, 'remission_rel_risk': rel_risk})

    markov_rel_risk = pd.DataFrame(c)
    return markov_rel_risk.pivot(
        index='exposure', columns='event', values='remission_rel_risk'
    )


# Print ################################################################################################################
def print_counts(pot):
    df = pot.tb.copy()
    df['count'] = df['pval']
    del df['pval']
    print(df)


def print_relrisk_and_table(pstore, event, exposure):
    pot = pstore[exposure, event]
    rr = pot.relative_risk(event, exposure, smooth_count=1)
    print(f"Relative risk of {event} when {exposure}:\n\t{rr:0.2f}")
    print("Based on the count table:")
    print_counts(pot)


def print_relrisk_and_table_from_pot(pot, event, exposure):
    rr = pot.relative_risk(event, exposure, smooth_count=1)
    print(f"Relative risk of {event} when {exposure}:\n\t{rr:0.2f}")
    print("Based on the count table:")
    print_counts(pot)


def print_remission_influence_supporting_info(pstore, event, exposure):
    """
    Prints remission influence supporting information.
    By "remission influence" we mean the relative risk of the event, given the exposure to an influencer
    during the "previous" time tick, conditioned on the existence of the event.
    That is, we're looking at:
        P(event=true | exposure-1=true, event-1=true)
    divided by
        P(event=true | exposure-1=false, event-1=true)
    """
    if isinstance(event, str):
        event = VarStr(event)
    if isinstance(exposure, str):
        exposure = VarStr(exposure)
    p = pstore[exposure - 1, event - 1, event]
    print(p)
    event_last_year = p * Pot.from_hard_evidence(**{event - 1: 1}) >> [
        exposure - 1,
        event,
    ]
    print_relrisk_and_table_from_pot(event_last_year, event, exposure - 1)


# Plot #################################################################################################################


def get_tick_and_labels(y_ticks, y_tick_labels=None):
    y_tick = []
    y_tick_label = []
    for y in y_ticks:
        if int(y) == y:
            y_tick.append(y)
            if y == 0:
                y_tick_label.append("")
            else:
                y_tick_label.append(str(np.sign(y) * int(2 ** abs(y))) + 'x')
    return y_tick, y_tick_label


def format_for_influencer_plot():
    t, _ = plt.yticks()
    if max(abs(t)) > 1:
        plt.yticks(*get_tick_and_labels(t))
    else:
        plt.yticks(t, list(map(lambda x: "{:0.2f}".format(2**x), t)))
    plt.grid(axis='y')


def diagonal_rr(lrr):
    c = list()
    for i, j in zip(lrr.index, lrr.columns):
        assert i.startswith(j)
        c.append({'x': i, 'var': j, 'val': lrr.loc[i, j]})
    return pd.DataFrame(c)[['var', 'val']].set_index('var')['val']


def plot_diagonal(lrr, figsize=dflt_figsize, **kwargs):
    t = diagonal_rr(lrr)
    kwargs = dict(title='Self (log2) relative risk', figsize=figsize, **kwargs)
    t.plot(kind='bar', **kwargs)
    format_for_influencer_plot()


def plot_influencers(lrr, var, figsize=dflt_figsize, **kwargs):
    kwargs = dict(
        title='Log2 relative risks for event: {}'.format(var), figsize=figsize, **kwargs
    )
    lrr.loc[:, var].plot(kind='bar', **kwargs)
    format_for_influencer_plot()


def plot_influenced(lrr, var, figsize=dflt_figsize, **kwargs):
    kwargs = dict(
        title='Log2 relative risks when exposed to {}'.format(var),
        figsize=figsize,
        **kwargs,
    )
    lrr.loc[var - 1, :].plot(kind='bar', **kwargs)
    format_for_influencer_plot()


def plot_remission_influencers(lrr, var, figsize=dflt_figsize, **kwargs):
    kwargs = dict(
        title='Log2 remission relative risks for event: {}'.format(var),
        figsize=figsize,
        **kwargs,
    )
    lrr.loc[:, var].plot(kind='bar', **kwargs)
    format_for_influencer_plot()


def plot_remission_influenced(lrr, var, figsize=dflt_figsize, **kwargs):
    kwargs = dict(
        title='Log2 remission relative risks when exposed to {}'.format(var),
        figsize=figsize,
        **kwargs,
    )
    lrr.loc[var, :].plot(kind='bar', **kwargs)
    format_for_influencer_plot()



from functools import partial
import os

import pandas as pd
import numpy as np
from openpyxl import load_workbook  # Get it here: https://pypi.org/project/openpyxl/ (or pip install openpyxl)


def pra(obj):
    print(*(x for x in dir(obj) if not x[0].startswith('_') and not x[0].isupper()), sep='\t')


def last_true_idx(series):
    return len(series) - series.iloc[::-1].values.argmax()


def mk_dotpath_getter(dotpath):
    attrs = dotpath.split('.')

    def attr_get(x):
        for a in attrs:
            x = getattr(x, a)
        return x

    return attr_get


def dfof(sheet, dotpath='value'):
    attr_get = mk_dotpath_getter(dotpath)
    data = [[attr_get(c) for c in row] for row in sheet]
    return pd.DataFrame(data)


def coord_val_and_bgcolor_of_cell(c):
    return {'coordinate': c.coordinate, 'value': c.value, 'rgb': c.fill.bgColor.rgb}


def crop(df, last_row_idx, last_col_idx):
    return df.iloc[:last_row_idx, :last_col_idx]


def matmask(df, mask):
    if isinstance(mask, pd.DataFrame):
        mask = mask.values
    d = df.values.copy()
    d[~mask] = ''
    return pd.DataFrame(d, index=df.index, columns=df.columns)


def row_lidx_to_mat_lidx(row_lidx, n_cols):
    return np.tile(row_lidx, (n_cols, 1)).T


white = '00000000'


def valid_vals_int_coords(vals_df):
    first_column = vals_df.iloc[:, 0]
    last_row_idx = last_true_idx((first_column != 'List of States') & (~first_column.isna()))

    age_row = vals_df.iloc[2, :]
    last_col_idx = last_true_idx(~age_row.isna())

    return last_row_idx, last_col_idx


def sheet_to_format_prepped_df(sheet):
    vals = dfof(sheet, dotpath='value')
    color = dfof(sheet, dotpath='fill.bgColor.rgb')
    coordinate = dfof(sheet, dotpath='coordinate')
    #     print(vals.shape, color.shape, coordinate.shape)

    int_coords = valid_vals_int_coords(vals)
    vals, color, coordinate = list(map(lambda x: crop(x, *int_coords), (vals, color, coordinate)))
    #     print(vals.shape, color.shape, coordinate.shape)

    df = pd.concat((vals, color, coordinate), axis=1, keys=('vals', 'color', 'coordinate'))
    #     df = df[~df.iloc[:, 0].isna()]

    return df


def lidx_of_empty_non_white_or_non_empty_white(df):
    is_null = df.vals.isnull()
    is_white = df.color == white
    return (is_null & ~is_white) | (~is_null & is_white)


def mk_known_exceptions_lidx(vals_df):
    n_rows, n_cols = vals_df.shape
    mk_mat_lidx = partial(row_lidx_to_mat_lidx, n_cols=n_cols)

    lidx = mk_mat_lidx([False] * 3 + [True] * (n_rows - 3))
    lidx[:, 0] = False
    lidx &= mk_mat_lidx(~vals_df.iloc[:, 0].isna())
    lidx &= mk_mat_lidx(~(vals_df.iloc[:, 0] == 'Select States below'))
    return lidx


def diagnosis_lidx(df):
    lidx = lidx_of_empty_non_white_or_non_empty_white(df)
    lidx &= mk_known_exceptions_lidx(df.vals)
    return lidx


def diagnosis_lidx_of_xls_file(xls_filepath, sheetname=None):
    wb = load_workbook(xls_filepath, data_only=True)
    sheetname = sheetname or next(iter(wb.sheetnames))
    sheet = wb[sheetname]

    df = sheet_to_format_prepped_df(sheet)
    lidx = lidx_of_empty_non_white_or_non_empty_white(df)
    lidx &= mk_known_exceptions_lidx(df.vals)
    return df, lidx


def diagnosis_items(source_dir, sheetname=None):
    is_excel = lambda x: x.endswith('.xlsx')
    join_dir = lambda *p: os.path.join(source_dir, *p)
    for xls_filepath in map(join_dir, filter(is_excel, os.listdir(source_dir))):
        filename = os.path.basename(xls_filepath)
        df, lidx = diagnosis_lidx_of_xls_file(xls_filepath, sheetname)
        if lidx.sum().sum() != 0:
            yield filename, df, lidx


def print_diagnosis(source_dir='.', sheetname=None):
    """Print the diagnosis of the raw xls files in `source_dir`.
    Namely, for every file that has a problem, the script will print the name of the problematic files,
    and print the coordinates of the cells that should be checked for problems.

    The problems that are checked for:
    - Cells (in the "data" range) that are (a) empty, but have color or (b) NOT empty, but without color.

    See functions `lidx_of_empty_non_white_or_non_empty_white` and `mk_known_exceptions_lidx` functions for details.
    """
    source_dir = os.path.abspath(os.path.expanduser(source_dir))
    print(f'Diagnosing xls files of "{source_dir}" ...\n\n')

    for filename, df, lidx in diagnosis_items(source_dir, sheetname):
        print(f"The possibly problematic cells for {filename}:")
        t = np.ravel(matmask(df.coordinate, lidx))
        t = t[t != '']
        print(*t, sep=', ')
        print("\n")


if __name__ == '__main__':
    import argh

    argh.dispatch_command(print_diagnosis)



from collections import defaultdict, Counter
import matplotlib.pylab as plt
import numpy as np
import pandas as pd
from py2store.caching import mk_memoizer  # like from cachetools import cached
import os
import re
from io import BytesIO

# get py2store here: https://github.com/i2mint/py2store
from py2store.base import Store
from py2store.stores.local_store import RelativePathFormatStore, LocalBinaryStore
from py2store.mixins import ReadOnlyMixin

from odus.nothing import nothing
from spyn.ppi.pot import Pot

from odus.sequential_var_sets import PVar, VarSet, DfData, extract_kps
from odus.plot_utils import heatmap

# from ut.pplot.matrix import heatmap

path_sep = os.path.sep


def ensure_slash_suffix(x):
    if not x.endswith(path_sep):
        return x + path_sep
    return x


def _commun_columns_of_dfs(dfs):
    categories = nothing
    df = None
    for df in dfs:
        categories = categories.intersection(set(df.columns))
    if df is not None:
        categories = [ss for ss in df.columns if ss in categories]
        return categories
    else:
        return []


def _all_columns_of_dfs(dfs):
    categories = []
    for df in dfs:
        for c in df.columns:
            if c not in categories:
                categories.append(c)
    return categories


simple_cat_p = re.compile('\W')


class Struct:
    def __init__(self, **kwargs):
        for attr, val in kwargs.items():
            setattr(self, attr, val)


class VarStr(str):
    def __add__(self, other):
        if isinstance(other, int):
            other = '+' + str(other)
        return super().__add__(other)

    def __sub__(self, other):
        if isinstance(other, int):
            other = '-' + str(other)
        return super().__add__(other)


def two_way_map(keys, val_of_key_func):
    val_of_key = dict()
    key_of_val = dict()
    for k in keys:
        v = val_of_key_func(k)
        val_of_key[k] = v
        key_of_val[v] = k
    return val_of_key, key_of_val


def mk_pvar_struct(df_store, only_for_cols_in_all_dfs=False):
    if only_for_cols_in_all_dfs:
        categories = _commun_columns_of_dfs(df_store.values())
    else:
        categories = _all_columns_of_dfs(df_store.values())
    val_of_attr = {
        simple_cat_p.sub('_', c.lower().strip()): PVar(c) for c in categories
    }
    return Struct(**val_of_attr)


def mk_pvar_str_struct(df_store, only_for_cols_in_all_dfs=False):
    if isinstance(df_store, Struct):
        return Struct(**{k: VarStr(v) for k, v in df_store.__dict__.items()})
    else:
        return mk_pvar_str_struct(
            mk_pvar_str_struct(df_store, only_for_cols_in_all_dfs)
        )


class HashableMixin:
    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return hash(self) == hash(other)


class DfStore(HashableMixin, ReadOnlyMixin, LocalBinaryStore):

    @mk_memoizer({})
    def __getitem__(self, k):
        return super().__getitem__(k)

    def _obj_of_data(self, data):
        df = pd.read_excel(BytesIO(data), header=0)
        df = df.iloc[:, 1:]
        df.iloc[0, 0] = 'year'
        df.iloc[1, 0] = 'age'
        df = df.rename(columns={'INTERVIEW ID NUMBER (enter in red)': 'category'})
        df = df.T
        df.columns = df.iloc[0, :]
        df = df.iloc[1:, 1:]  # TODO: would like to keep year, but not in column
        df = df.set_index('age')
        df = df.astype(int)
        df[df != 0] = (
            1  # TODO: May want to remove this to expose mistakes instead of repairing them
        )
        return df


class VarSetCountsStore(HashableMixin, ReadOnlyMixin, Store):
    def __init__(self, df_store, store=None):
        self.df_store = df_store
        super().__init__(store)

    def mk_pvar_attrs(self, only_for_cols_in_all_dfs=False):
        self.v = mk_pvar_struct(self.df_store, only_for_cols_in_all_dfs)

    @mk_memoizer({})
    def __getitem__(self, k):
        if isinstance(k, (tuple, PVar)):
            return self.__getitem__(VarSet(k))
        else:
            c = Counter()
            for df in self.df_store.values():
                # counts = DfData(df).extract_kps(k)
                counts = extract_kps(df, k)
                c.update(counts)
            return c


class PotStore(VarSetCountsStore):
    @mk_memoizer({})
    def __getitem__(self, k):
        if isinstance(k, (tuple, PVar)):
            return self.__getitem__(VarSet(k))
        else:
            counter = super().__getitem__(k)
            d = list()
            for key, val in counter.items():
                dd = {str(k.varset[i]): var_val for i, var_val in enumerate(key)}
                dd['pval'] = val
                d.append(dd)
            # return pd.DataFrame(d)
            return Pot.from_count_df_to_count(
                pd.DataFrame(d)[k.varset_strs + ['pval']], count_col='pval'
            )


class DelegMap:
    def __init__(self, dfunc, dmap=None):
        if dmap is None:
            dmap = dict()
        self.dmap = dmap
        self.dfunc = dfunc

    def __getitem__(self, k):
        if k not in self.dmap:
            self.dmap[k] = self.dfunc(k)
        return self.dmap[k]


class Dacc:
    def __init__(self, xls_rootdir, categories=None):
        self.s = DfStore(ensure_slash_suffix(xls_rootdir) + '{}.xlsx', mode='b')
        if categories is None:
            categories = _commun_columns_of_dfs(self.s.values())
        self.categories = categories
        self.counts_of = defaultdict(Counter)

    def mk_counts_of_kps(self, kps_list):
        for f in self.s.keys():
            df = self.s[f][self.categories]
            dd = DfData(df)
            for k, v in dd.extract_with_key_pattern_sets(kps_list):
                self.counts_of[k].update([v])

    # def counts_of(self):


def counts_of_kps(store, categories, kps_list):
    counts_of = defaultdict(Counter)
    for f in store.keys():
        df = store[f][categories]
        dd = DfData(df)
        for k, v in dd.extract_with_key_pattern_sets(kps_list):
            counts_of[k].update([v])
    return counts_of


# vs = tuple(map(str, VarSet.from_str(s)))
# d = list()
# for k, v in t.items():
#     dd = {vs[i]: kk for i, kk in enumerate(k)}
#     dd['pval'] = v
#     d.append(dd)
# u = ProbPot.from_count_df_to_count(pd.DataFrame(d), count_col='pval')
# u

if __name__ == '__main__':
    import os

    proj_root = os.path.dirname(__file__)
    pjoin = lambda f: os.path.join(proj_root, f)
    xls_rootdir = pjoin('data/surveys/')

    # vs = tuple(map(str, VarSet.from_str(s)))

    # d = list()
    # for k, v in counts_of.items():
    #     dd = {vs[i]: kk for i, kk in enumerate(k)}
    #     dd['pval'] = v
    #     d.append(dd)
    # u = ProbPot.from_count_df_to_count(pd.DataFrame(d), count_col='pval')
    # u
